import operator
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.http import HttpResponseNotFound, HttpResponseForbidden
from django_yaaac.shortcuts import json_response
from django_yaaac.utils import lookup_dict_from_url_params


def cache_model(f):
    def wrapper(manager, request, *args, **kwargs):
        app = kwargs["app"]
        model = kwargs["model"]
        kwargs["model"] = ContentType.objects.get(app_label=app,
                                                  model=model).model_class()
        return f(manager, request, *args, **kwargs)
    return wrapper


def check_permissions(f):
    def wrapper(manager, request, *args, **kwargs):
        model = kwargs["model"]
        suggest_by = request.GET.get('suggest_by')
        try:
            options = model.Yaaac()
            options.user_passes_test
            options.allows_suggest_by
        except AttributeError:
            return HttpResponseNotFound()
        if not options.user_passes_test(request.user):
            return HttpResponseForbidden()
        if not suggest_by and not request.GET.get("pk"):
            return HttpResponseNotFound()
        if suggest_by and suggest_by not in options.allows_suggest_by:
            return HttpResponseForbidden()

        return f(manager, request, *args, **kwargs)
    return wrapper


class AutocompleteManager(object):
    def __init__(self, name="yaaac", app_name="yaaac"):
        self.name = name
        self.app_name = app_name

    def get_urls(self):
        from django.conf.urls import patterns, url
        urlpatterns = patterns('',
            url(r'^(?P<app>\w+)/(?P<model>\w+)/search/$', self.search, name='search'),
        )
        return urlpatterns

    @property
    def urls(self):
        return self.get_urls(), self.app_name, self.name

    @cache_model
    @check_permissions
    def search(self, request, app, model):
        query = request.GET.get('query')
        pk = request.GET.get('pk')
        if pk:
            obj = model.objects.get(pk=pk)
            try:
                url = obj.get_absolute_url()
            except AttributeError:
                url = None
            return json_response({
                "value": unicode(obj),
                "url": url,
            })
        search_fields = request.GET.get('search_fields').split(",")
        suggest_by = request.GET.get('suggest_by')

        filter_params = request.GET.copy()
        del filter_params["t"]
        del filter_params["query"]
        del filter_params["search_fields"]
        if "suggest_by" in filter_params:
            del filter_params["suggest_by"]
        kwargs = lookup_dict_from_url_params(filter_params)
        result = model.objects.filter(**kwargs)
        result = self.get_search_results(result, search_fields, query)

        if suggest_by in model._meta.get_all_field_names():
            result = result.values_list('id', suggest_by)
        else:
            result = [(obj.pk, getattr(obj, suggest_by)()) for obj in result]
        result = result or [('', '')]
        suggestions = [{"value": r[1], "data": r[0]} for r in result]
        return json_response({'query': request.GET.get('query'),
                              'suggestions': suggestions})

    # From https://github.com/django/django/blob/master/django/contrib/admin/options.py
    # without the use_distinct calculation.
    def get_search_results(self, queryset, search_fields, search_term):
        def construct_search(field_name):
            if field_name.startswith('^'):
                return "%s__istartswith" % field_name[1:]
            elif field_name.startswith('='):
                return "%s__iexact" % field_name[1:]
            elif field_name.startswith('@'):
                return "%s__search" % field_name[1:]
            else:
                return "%s__icontains" % field_name

        orm_lookups = [construct_search(str(search_field))
                       for search_field in search_fields]
        for bit in search_term.split():
            or_queries = [models.Q(**{orm_lookup: bit})
                          for orm_lookup in orm_lookups]
            queryset = queryset.filter(reduce(operator.or_, or_queries))
        return queryset


autocomplete = AutocompleteManager()



import json
from django import VERSION as DJ_VERSION
from django import forms
from django.contrib.admin.templatetags.admin_static import static
from django.core.urlresolvers import reverse_lazy
if DJ_VERSION < (1, 7):
    from django.forms.util import flatatt
else:
    from django.forms.utils import flatatt
from django.utils.html import format_html
from django.utils.text import Truncator
from django_yaaac.utils import clean_fieldname_prefix



JS_COMPAT_FILE = "yaaac_compat.js"
if DJ_VERSION < (1, 6):
    JS_COMPAT_FILE = "yaaac_compat_legacy.js"


class AutocompleteWidget(forms.HiddenInput):
    is_hidden = False

    class Media:
        css = {
            'all': (
                static('django_yaaac/css/autocomplete.css'),
            )
        }
        js = (
            static('django_yaaac/js/jquery.autocomplete.min.js'),
            static('django_yaaac/js/%s' % JS_COMPAT_FILE),
            static('django_yaaac/js/yaaac_autocomplete.js'),
        )

    def __init__(self, site, model, limit_choices_to=None, opts=None, attrs=None):
        self.site = site
        self.model = model
        self.opts = {
            "min_chars": 1,
            "max_height": 300,
            "width": 300,
            "suggest_by": "__unicode__",
        }
        # opts should not by None.
        self.opts.update(opts)

        self.limit_choices_to = limit_choices_to or {}
        super(AutocompleteWidget, self).__init__(attrs)

    def render(self, name, value, attrs=None):
        app_label = self.model._meta.app_label
        if DJ_VERSION < (1, 6):
            model_name = self.model._meta.module_name
        else:
            model_name = self.model._meta.model_name
        info = (self.site.name, app_label, model_name)
        search_url = reverse_lazy("yaaac:search", kwargs={"app": app_label, "model": model_name})
        
        # Cannot just do self.opts.pop("search_fields") because render() must be side-effect free.
        search_fields = self.opts.get("search_fields")
        search_opts = self.opts.copy()
        del search_opts["search_fields"]

        lookup_url = reverse_lazy('%s:%s_%s_changelist' % info, current_app=self.site.name)
        params = self.url_parameters()
        url_params = '?' + '&'.join('%s=%s' % (k, v) for k, v in params.items())

        attrs.update({
            'class': 'yaaac_%s yaaac_pk vForeignKeyRawIdAdminField' % clean_fieldname_prefix(name),
            'search_url': "%s%s&suggest_by=%s" % (search_url, url_params, self.opts["suggest_by"]),
            'search_opts': json.dumps(search_opts),
            'search_fields': ",".join(search_fields), 
        })
        hidden_input = super(AutocompleteWidget, self).render(name, value, attrs)
        autocomp_input = format_html('<input{0} />',
                                     flatatt({"type": "text",
                                              "class": "yaaac_search_input",
                                              "placeholder": "start typing to search",
                                              "style": value and "display:none" or ""}))
        lookup_elem = format_html('<a {0}><img {1} /></a>',
                                  flatatt({"href": "%s%s" % (lookup_url, url_params),
                                           "id": "lookup_id_%s" % name,
                                           "class": "yaaac_lookup",
                                          "style": value and "display:none" or ""}),
                                  flatatt({"width": "16", "height": "16", "alt": "Lookup",
                                           "src": static('django_yaaac/img/selector-search.gif')}))
        
        return format_html(u'<span class="yaaac_container">{0}{1}{2}{3}</span>',
                           hidden_input, autocomp_input, lookup_elem, self.value_elem(value))

    def value_elem(self, value):
        style = 'display:none'
        label = hasattr(self.model, "get_absolute_url") and format_html('<a></a>') or ''
        clear_elem = format_html('<span class="yaaac_clear_value">x</span>')
        if value:
            obj = self.model._default_manager.get(pk=value)
            label = Truncator(obj).words(14, truncate=" ...")
            style = ''
            try:
                url = obj.get_absolute_url()
            except AttributeError:
                pass
            else:
                label = format_html(u'<a{0}>{1}</a>', flatatt({"href": url}), label)
        label_elem = format_html(u'<span{0}>{1}</span>', flatatt({"class": "yaaac_value"}), label)
        return format_html(u'<span{0}>{1}{2}</span>',
                           flatatt({"class": "yaaac_value_container", "style": style}),
                           label_elem, clear_elem)

    def base_url_parameters(self):
        from django.contrib.admin.widgets import url_params_from_lookup_dict
        return url_params_from_lookup_dict(self.limit_choices_to)

    def url_parameters(self):
        from django.contrib.admin.views.main import TO_FIELD_VAR
        params = self.base_url_parameters()
        params.update({TO_FIELD_VAR: "id"}) # Hardcoded here because we do not have 'rel' object.
        return params



from django.forms.models import ModelChoiceField
from django_yaaac.forms.widgets import AutocompleteWidget


class AutocompleteModelChoiceField(ModelChoiceField):
    widget = AutocompleteWidget

    def __init__(self, site, queryset, yaaac_opts, limit_choices_to=None,
                 empty_label="---------", cache_choices=False, required=True,
                 widget=None, label=None, initial=None, help_text='',
                 to_field_name=None, *args, **kwargs):
        # the `limit_choices_to' parameter allows us to enable the filtering in
        # the ajax autocomplete and the related lookup.
        # It may seems redundant with `ForeignKey.limit_choices_to' but we don't 
        # have access to a `rel' object here.
        model = queryset.model
        widget = AutocompleteWidget(site, model, limit_choices_to, yaaac_opts)
        ModelChoiceField.__init__(self, queryset, empty_label, cache_choices,
                                  required, widget, label, initial, help_text,
                                  to_field_name, *args, **kwargs)



import os
from distutils.core import setup

README = open(os.path.join(os.path.dirname(__file__), 'README.txt')).read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-yaaac',
    version='0.93.0',
    license='BSD Licence',
    author='Stanislas Guerra',
    author_email='stanislas.guerra@gmail.com',
    description='Django application providing Ajax search capabilities.',
    long_description = README,
    url='https://github.com/Starou/django-yaaac',
    packages=['django_yaaac',
              'django_yaaac.forms'],
    package_data={
        'django_yaaac': [
            'static/django_yaaac/css/*.css',
            'static/django_yaaac/js/*.js',
            'static/django_yaaac/img/*.png',
            'static/django_yaaac/img/*.jpg',
            'static/django_yaaac/img/*.gif',
        ]
    },
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



def lookup_dict_from_url_params(url_params):
    """Convert a url_params dict in a mapping suitable for ORM querying."""
    lookup_dict = {}
    for k, v in url_params.items():
        values = v.split(",")
        if len(values) > 1:
            lookup_dict[k] = values
        else:
            lookup_dict[k] = v
    return lookup_dict

def clean_fieldname_prefix(fieldname):
    import re
    try:
        fieldname = re.match(r'(.*)-(\w+)', fieldname).groups()[1]
    except AttributeError:
        pass
    return fieldname
